from multipledispatch.variadic import isvariadic, Variadic


class A(object): pass
class B(A): pass
class C(object): pass


def test_is_variadic():
    assert isvariadic(Variadic[A])
    assert not isvariadic(A)


def test_equals_simple():
    type_a = Variadic[A]
    type_b = Variadic[B]
    other_type_a = Variadic[A]
    assert type_a == other_type_a
    assert not type_a == type_b


def test_equals_union():
    union_a_b = Variadic[(A, B)]
    union_b_a = Variadic[(B, A)]

    union_a_b_c = Variadic[(A, B, C)]
    union_c_b_a = Variadic[(C, B, A)]

    assert union_a_b == union_b_a
    assert union_a_b_c == union_c_b_a

    assert not union_a_b == union_a_b_c



import six

from multipledispatch import dispatch


@dispatch(int)
def isint(x):
    return True


@dispatch(object)
def isint(x):
    return False


def test_simple():
    for i in six.moves.xrange(100000):
        assert isint(5)
        assert not isint('a')



from multipledispatch.conflict import (supercedes, ordering, ambiguities,
        ambiguous, super_signature, consistent)
from multipledispatch.dispatcher import Variadic


class A(object): pass
class B(A): pass
class C(object): pass

def _test_supercedes(a, b):
    assert supercedes(a, b)
    assert not supercedes(b, a)


def test_supercedes():
    _test_supercedes([B], [A])
    _test_supercedes([B, A], [A, A])


def test_consistent():
    assert consistent([A], [A])
    assert consistent([B], [B])
    assert not consistent([A], [C])
    assert consistent([A, B], [A, B])
    assert consistent([B, A], [A, B])
    assert not consistent([B, A], [B])
    assert not consistent([B, A], [B, C])


def test_super_signature():
    assert super_signature([[A]]) == [A]
    assert super_signature([[A], [B]]) == [B]
    assert super_signature([[A, B], [B, A]]) == [B, B]
    assert super_signature([[A, A, B], [A, B, A], [B, A, A]]) == [B, B, B]


def test_ambiguous():
    assert not ambiguous([A], [A])
    assert not ambiguous([A], [B])
    assert not ambiguous([B], [B])
    assert not ambiguous([A, B], [B, B])
    assert ambiguous([A, B], [B, A])


def test_ambiguities():
    signatures = [[A], [B], [A, B], [B, A], [A, C]]
    expected = set([((A, B), (B, A))])
    result = ambiguities(signatures)
    assert set(map(frozenset, expected)) == set(map(frozenset, result))

    signatures = [[A], [B], [A, B], [B, A], [A, C], [B, B]]
    expected = set()
    result = ambiguities(signatures)
    assert set(map(frozenset, expected)) == set(map(frozenset, result))


def test_ordering():
    signatures = [[A, A], [A, B], [B, A], [B, B], [A, C]]
    ord = ordering(signatures)
    assert ord[0] == (B, B) or ord[0] == (A, C)
    assert ord[-1] == (A, A) or ord[-1] == (A, C)


def test_type_mro():
    assert super_signature([[object], [type]]) == [type]


def test_supercedes_variadic():
    _test_supercedes((Variadic[B],), (Variadic[A],))
    _test_supercedes((B, Variadic[A]), (Variadic[A],))
    _test_supercedes((Variadic[A],), (Variadic[(A, C)],))
    _test_supercedes((A, B, Variadic[C]), (Variadic[object],))
    _test_supercedes((A, Variadic[B]), (Variadic[A],))
    _test_supercedes(tuple([]), (Variadic[A],))
    _test_supercedes((A, A, A), (A, Variadic[A]))


def test_consistent_variadic():
    # basic check
    assert consistent((Variadic[A],), (Variadic[A],))
    assert consistent((Variadic[B],), (Variadic[B],))
    assert not consistent((Variadic[C],), (Variadic[A],))

    # union types
    assert consistent((Variadic[(A, C)],), (Variadic[A],))
    assert consistent((Variadic[(A, C)],), (Variadic[(C, A)],))
    assert consistent((Variadic[(A, B, C)],), (Variadic[(C, B, A)],))
    assert consistent((A, B, C), (Variadic[(A, B, C)],))
    assert consistent((A, B, C), (A, Variadic[(B, C)]))

    # more complex examples
    assert consistent(tuple([]), (Variadic[object],))
    assert consistent((A, A, B), (A, A, Variadic[B]))
    assert consistent((A, A, B), (A, A, Variadic[A]))
    assert consistent((A, B, Variadic[C]), (B, A, Variadic[C]))
    assert consistent((A, B, Variadic[C]), (B, A, Variadic[(C, B)]))

    # not consistent
    assert not consistent((C,), (Variadic[A],))
    assert not consistent((A, A, Variadic[C]), (A, Variadic[C]))
    assert not consistent((A, B, Variadic[C]), (C, B, Variadic[C]))



from multipledispatch import dispatch
import pytest


@dispatch(int)
def isint(x):
    return True


@dispatch(object)
def isint(x):
    return False


@dispatch(object, object)
def isint(x, y):
    return False


@pytest.mark.parametrize("val", [1, 'a'])
def test_benchmark_call_single_dispatch(benchmark, val):
    benchmark(isint, val)


@pytest.mark.parametrize("val", [(1, 4)])
def test_benchmark_call_multiple_dispatch(benchmark, val):
    benchmark(isint, *val)


def test_benchmark_add_and_use_instance(benchmark):
    namespace = {}

    @benchmark
    def inner():
        @dispatch(int, int, namespace=namespace)
        def mul(x, y):
            return x * y

        @dispatch(str, int, namespace=namespace)
        def mul(x, y):
            return x * y

        @dispatch(int, int, [float], namespace=namespace)
        def mul(x, y, *args):
            return x * y

        @dispatch([int], namespace=namespace)
        def mul(*args):
            return sum(args)

        mul(4, 5)
        mul('x', 5)
        mul(1, 2, 3., 4., 5.)
        mul(1, 2, 3, 4, 5)




import warnings

from multipledispatch.dispatcher import (Dispatcher, MDNotImplementedError,
                                         MethodDispatcher)
from multipledispatch.conflict import ambiguities
from multipledispatch.utils import raises


def identity(x):
    return x


def inc(x):
    return x + 1


def dec(x):
    return x - 1


def test_dispatcher():
    f = Dispatcher('f')
    f.add((int,), inc)
    f.add((float,), dec)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert f.resolve((int,)) == inc
    assert f.dispatch(int) is inc

    assert f(1) == 2
    assert f(1.0) == 0.0


def test_union_types():
    f = Dispatcher('f')
    f.register((int, float))(inc)

    assert f(1) == 2
    assert f(1.0) == 2.0


def test_dispatcher_as_decorator():
    f = Dispatcher('f')

    @f.register(int)
    def inc(x):
        return x + 1

    @f.register(float)
    def dec(x):
        return x - 1

    assert f(1) == 2
    assert f(1.0) == 0.0


def test_register_instance_method():

    class Test(object):
        __init__ = MethodDispatcher('f')

        @__init__.register(list)
        def _init_list(self, data):
            self.data = data

        @__init__.register(object)
        def _init_obj(self, datum):
            self.data = [datum]

    a = Test(3)
    b = Test([3])
    assert a.data == b.data


def test_on_ambiguity():
    f = Dispatcher('f')

    def identity(x):
        return x

    ambiguities = [False]

    def on_ambiguity(dispatcher, amb):
        ambiguities[0] = True

    f.add((object, object), identity)
    f.add((object, float), identity)
    f.add((float, object), identity)

    assert not ambiguities[0]
    f.reorder(on_ambiguity=on_ambiguity)
    assert ambiguities[0]


def test_serializable():
    f = Dispatcher('f')
    f.add((int,), inc)
    f.add((float,), dec)
    f.add((object,), identity)

    import pickle
    assert isinstance(pickle.dumps(f), (str, bytes))

    g = pickle.loads(pickle.dumps(f))

    assert g(1) == 2
    assert g(1.0) == 0.0
    assert g('hello') == 'hello'


def test_raise_error_on_non_class():
    f = Dispatcher('f')
    assert raises(TypeError, lambda: f.add((1,), inc))


def test_docstring():

    def one(x, y):
        """ Docstring number one """
        return x + y

    def two(x, y):
        """ Docstring number two """
        return x + y

    def three(x, y):
        return x + y

    master_doc = 'Doc of the multimethod itself'

    f = Dispatcher('f', doc=master_doc)
    f.add((object, object), one)
    f.add((int, int), two)
    f.add((float, float), three)

    assert one.__doc__.strip() in f.__doc__
    assert two.__doc__.strip() in f.__doc__
    assert (
        f.__doc__.find(one.__doc__.strip()) <
        f.__doc__.find(two.__doc__.strip())
    )
    assert 'object, object' in f.__doc__
    assert master_doc in f.__doc__


def test_help():
    def one(x, y):
        """ Docstring number one """
        return x + y

    def two(x, y):
        """ Docstring number two """
        return x + y

    def three(x, y):
        """ Docstring number three """
        return x + y

    master_doc = 'Doc of the multimethod itself'

    f = Dispatcher('f', doc=master_doc)
    f.add((object, object), one)
    f.add((int, int), two)
    f.add((float, float), three)

    assert f._help(1, 1) == two.__doc__
    assert f._help(1.0, 2.0) == three.__doc__


def test_source():
    def one(x, y):
        """ Docstring number one """
        return x + y

    def two(x, y):
        """ Docstring number two """
        return x - y

    master_doc = 'Doc of the multimethod itself'

    f = Dispatcher('f', doc=master_doc)
    f.add((int, int), one)
    f.add((float, float), two)

    assert 'x + y' in f._source(1, 1)
    assert 'x - y' in f._source(1.0, 1.0)


def test_source_raises_on_missing_function():
    f = Dispatcher('f')

    assert raises(TypeError, lambda: f.source(1))


def test_no_implementations():
    f = Dispatcher('f')
    assert raises(NotImplementedError, lambda: f('hello'))


def test_register_stacking():
    f = Dispatcher('f')

    @f.register(list)
    @f.register(tuple)
    def rev(x):
        return x[::-1]

    assert f((1, 2, 3)) == (3, 2, 1)
    assert f([1, 2, 3]) == [3, 2, 1]

    assert raises(NotImplementedError, lambda: f('hello'))
    assert rev('hello') == 'olleh'


def test_dispatch_method():
    f = Dispatcher('f')

    @f.register(list)
    def rev(x):
        return x[::-1]

    @f.register(int, int)
    def add(x, y):
        return x + y

    class MyList(list):
        pass

    assert f.dispatch(list) is rev
    assert f.dispatch(MyList) is rev
    assert f.dispatch(int, int) is add


def test_not_implemented():
    f = Dispatcher('f')

    @f.register(object)
    def _1(x):
        return 'default'

    @f.register(int)
    def _2(x):
        if x % 2 == 0:
            return 'even'
        else:
            raise MDNotImplementedError()

    assert f('hello') == 'default'  # default behavior
    assert f(2) == 'even'           # specialized behavior
    assert f(3) == 'default'        # fall back to default behavior
    assert raises(NotImplementedError, lambda: f(1, 2))


def test_not_implemented_error():
    f = Dispatcher('f')

    @f.register(float)
    def _(a):
        raise MDNotImplementedError()

    assert raises(NotImplementedError, lambda: f(1.0))


def test_vararg_not_last_element_of_signature():
    f = Dispatcher('f')
    assert raises(TypeError, lambda: f.register([float], str)(lambda: None))


def test_vararg_has_multiple_elements():
    f = Dispatcher('f')
    assert raises(TypeError, lambda: f.register([float, str])(lambda: None))


def test_vararg_dispatch_simple():
    f = Dispatcher('f')

    @f.register([float])
    def _1(*args):
        return args

    assert f(1.0) == (1.0,)
    assert f(1.0, 2.0, 3.0) == (1.0, 2.0, 3.0)


def test_vararg_dispatch_ambiguity():
    f = Dispatcher('f')

    @f.register(float, object, [int])
    def _1(a, b, *args):
        return 1

    @f.register(object, float, [int])
    def _2(a, b, *args):
        return 2

    assert ambiguities(f.funcs)


def test_vararg_dispatch_ambiguity_in_variadic():
    f = Dispatcher('f')

    @f.register(float, [object])
    def _1(a, b, *args):
        return 1

    @f.register(object, [float])
    def _2(a, b, *args):
        return 2

    assert ambiguities(f.funcs)


def test_vararg_dispatch_multiple_types_explicit_args():
    f = Dispatcher('f')

    @f.register(str, [float])
    def _1(a, *b):
        return (a, b)

    result = f('a', 1.0, 2.0, 3.0)
    assert result == ('a', (1.0, 2.0, 3.0))


def test_vararg_dispatch_multiple_implementations():
    f = Dispatcher('f')

    @f.register(str, [float])
    def _1(a, *b):
        return 'mixed_string_floats'

    @f.register([float])
    def _2(*b):
        return 'floats'

    @f.register([str])
    def _3(*strings):
        return 'strings'

    assert f('a', 1.0, 2.0) == 'mixed_string_floats'
    assert f(1.0, 2.0, 3.14) == 'floats'
    assert f('a', 'b', 'c') == 'strings'


def test_vararg_dispatch_unions():
    f = Dispatcher('f')

    @f.register(str, [(int, float)])
    def _1(a, *b):
        return 'mixed_string_ints_floats'

    @f.register([str])
    def _2(*strings):
        return 'strings'

    @f.register([(str, int)])
    def _3(*strings_ints):
        return 'mixed_strings_ints'

    @f.register([object])
    def _4(*objects):
        return 'objects'

    assert f('a', 1.0, 7, 2.0, 11) == 'mixed_string_ints_floats'
    assert f('a', 'b', 'c') == 'strings'
    assert f('a', 1, 'b', 2) == 'mixed_strings_ints'
    assert f([], (), {}) == 'objects'


def test_vararg_no_args():
    f = Dispatcher('f')

    @f.register([str])
    def _1(*strings):
        return 'strings'

    assert f() == 'strings'


def test_vararg_no_args_failure():
    f = Dispatcher('f')

    @f.register(str, [str])
    def _2(*strings):
        return 'strings'

    assert raises(NotImplementedError, f)


def test_vararg_no_args_failure():
    f = Dispatcher('f')

    @f.register([str])
    def _2(*strings):
        return 'strings'

    assert raises(NotImplementedError, lambda: f('a', 'b', 1))


def test_vararg_ordering():
    f = Dispatcher('f')

    @f.register(str, int, [object])
    def _1(string, integer, *objects):
        return 1

    @f.register(str, [object])
    def _2(string, *objects):
        return 2

    @f.register([object])
    def _3(*objects):
        return 3

    assert f('a', 1) == 1
    assert f('a', 1, ['a']) == 1
    assert f('a', 1, 'a') == 1
    assert f('a', 'a') == 2
    assert f('a') == 2
    assert f('a', ['a']) == 2
    assert f(1) == 3
    assert f() == 3



#!/usr/bin/env python

from os.path import exists
from setuptools import setup
import multipledispatch

setup(name='multipledispatch',
      version=multipledispatch.__version__,
      description='Multiple dispatch',
      url='http://github.com/mrocklin/multipledispatch/',
      author='Matthew Rocklin',
      author_email='mrocklin@gmail.com',
      install_requires=['six'],
      license='BSD',
      keywords='dispatch',
      packages=['multipledispatch'],
      long_description=(open('README.rst').read() if exists('README.rst')
                        else ''),
      zip_safe=False)
